// BSP : STM32 boards groups by STMicroelectronics 2.7.1
// Borad select : STM Boards groups / LoRa boars
// Board part number : LoRa-E5 mini (Board to be selected from Tools submenu 'Board part number)

// 1. wait for transmission of Slave
// 2. receives Slave data(Slave UID, mills(), Vbatt)
// 3. transtits response data(Master UID, rssi, snr, interval)

// 2024/02/26

#include <Adafruit_GFX.h>       // 1.11.9
#include <Adafruit_SSD1306.h>   // 2.5.9
#include "myFonts/tahoma5pt7b.h"  // any font
#include "myFonts/tahoma7pt7b.h"
#include <RadioLib.h>           // 6.4.2  https://github.com/jgromes/RadioLib
#include <SPI.h>
#include <SD.h>

#define CHECK_INTERVAL 20000    // communication check interval [mS]

// BAT+_pin() --> 1000k --> A4_pin --> 510k --> D0_pin
#define PIN_VBATT      A4       // battery voltage measurement pin
#define READ_ENABLE    D0       // read PIN_VBATT enable LOW:enable, HIGH:disable

#define SD_CS         PB9       // SD chip select pin

// RM0461 Reference manual Rev 8  37.1.4 IEEE 64-bit unique device ID register (UID64)
#define MASTER_UID  *((uint32_t*)0x1FFF7580)    // use uid64 for Master device UID
//#define SLAVE_UID   0x05358291  // #1 Slave device UID
//#define SLAVE_UID   0x0535EE2F  // #2 Slave device UID
#define SLAVE_UID   0x05370378  // #3 Slave device UID

Adafruit_SSD1306 display(128, 64, &Wire, -1);
File logFile;

// \Arduino15\packages\STMicroelectronics\hardware\stm32\2.7.1
// \variants\STM32WLxx\WL54JCI_WL55JCI_WLE4J(8-B-C)I_WLE5J(8-B-C)I\variant_LORA_E5_MINI.h : 173
STM32WLx radio = new STM32WLx_Module();
static const uint32_t rfswitch_pins[] =
                         {PA4,  PA5,  RADIOLIB_NC};
static const Module::RfSwitchMode_t rfswitch_table[] = {
  {STM32WLx::MODE_IDLE,  {LOW,  LOW}},
  {STM32WLx::MODE_RX,    {HIGH, LOW}},
  {STM32WLx::MODE_TX_HP, {LOW, HIGH}},  // for LoRa-E5 mini
//  {STM32WLx::MODE_TX_LP, {HIGH, HIGH}},   // for LoRa-E5-LE mini
  END_OF_MODE_TABLE,
};

#define dataNum   12                // data bytes
union unionData {                   // for data type conversion
  uint32_t  dataBuff32[dataNum/4];
  float     dataBuff32F[dataNum/4];
  int16_t   dataBuff16[dataNum/2];
  uint8_t   dataBuff8[dataNum];
};
union unionData ud;

int rssi;                           // signal RSSI [dBm]
int snr;                            // signal SN ratio [dB]
uint32_t masteruid = MASTER_UID;    // Master device UID
uint32_t slaveuid;                  // Slave device UID
uint16_t sec;                       // mS/1000 data [sec]
float Vbatt;                        // battery voltage data [V]
char printBuff[4];                  // for sprintf()
String  txdata = "";                // transmission data packet string
String  rxdata = "";                // received data packet string
bool rxstatus = false;              // receive status flag
bool receivedFlag = false;          // flag that a packet was received
bool transmittedFlag = false;       // flag that a packet was transmitted
String logFileName = "LoRaLog.txt"; // Log file name
uint32_t timestamp = millis();      // previouse received time [mS]
uint32_t receivedTime = millis();   // packet received time [mS]
float receivedInterval;             // packet received interval [mS]

// *******************************************************************************************************
void setup() {
  Serial.begin(115200);
  while(!Serial);
  delay(2000);

  // Pin initialization
  pinMode(LED_RED, OUTPUT);         // built-in LED
  pinMode(READ_ENABLE, OUTPUT);     // read PIN_VBATT enable

  // Display initialization
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  display.clearDisplay();
  display.setFont(&tahoma7pt7b);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 15);
  display.println("LoRa-E5");
  display.setCursor(0, 31);
  display.println("P2P Test");
  display.setCursor(0, 47);
  display.println("Master");  
  display.display();
  delay(2000);
  display.clearDisplay();

  // SD initialization and file open
  if(!SD.begin(SD_CS)) {                     // SCK 4MHz   
    Serial.println("SD begin failure!");
    display.clearDisplay();
    display.setCursor(0, 15);
    display.println("SD begin failure");
    display.display();
    errorBlink(2);                           // error [[ 2 ]]
  } 

  logFile = SD.open(logFileName, FILE_WRITE);
  if(!logFile) { 
    Serial.println(logFileName + " open failure");
    display.clearDisplay();
    display.setCursor(0, 15);
    display.println("file open failure");
    display.display();
    errorBlink(3);                           // error [[ 3 ]]   
  }   
  Serial.println("SD initialization finished");

  // STM32WL initialization
  radio.setRfSwitchTable(rfswitch_pins, rfswitch_table);

  // int16_t begin(float freq = 434.0, float bw = 125.0, uint8_t sf = 9, uint8_t cr = 7, \
  // uint8_t syncWord = RADIOLIB_SX126X_SYNC_WORD_PRIVATE, int8_t power = 10, \
  // uint16_t preambleLength = 8, float tcxoVoltage = 1.6, bool useRegulatorLDO = false);
  int state = radio.begin(922.6, 125.0, 12, 5, 0x12, 13, 8, 1.7, 0);

  if (state == RADIOLIB_ERR_NONE) {
    Serial.println("radio.begin() success!");
  }
  else {
    Serial.print("failed, code ");
    Serial.println(state);
    errorBlink(4);                           // error [[ 4 ]]
  }

  // callback function when received or transmitted
  radio.setDio1Action(setFlag);

  // setup end 
  display.setCursor(0, 47);
  display.print("LoRa-E5 Master");
  display.setCursor(0, 63);
  display.print("Setup END");
  display.display();
  delay(2000);
  display.clearDisplay(); 
}

// ***********************************************************************************************************
void loop() 
{
  // Master receives packet from Slave ***********************************************************************
  Serial.println("Waiting for incoming Slave transmitting packet ... ");

  // start listening for LoRa packets
  receivedFlag = false;
  int state = radio.startReceive();
  if (state != RADIOLIB_ERR_NONE) {
    Serial.print("failed, code ");
    Serial.println(state);
    errorBlink_1(2);                           // error [[ 2 ]]
  }

  // check if the flag is set
  // If not received, proceed to next communication cycle to attempt synchronization recovery
  uint32_t checkstart = millis();
  digitalWrite(LED_RED, LOW);
  while(!receivedFlag & (millis() - checkstart < CHECK_INTERVAL)) {
    delay(1);   // wait for SPI ready
  }
  digitalWrite(LED_RED, HIGH);
  receivedFlag = false;

  // read a packet 
  state = radio.readData(ud.dataBuff8, dataNum);

  slaveuid = ud.dataBuff32[0];    // Slave device UID
  if (state == RADIOLIB_ERR_NONE & slaveuid == SLAVE_UID) {

    // packet form Slave was successfully received
    receivedTime = millis();
    receivedInterval = (receivedTime - timestamp) / 1000.0; // [sec]
    timestamp = receivedTime;

    rxdata = "";
    for(int i = 0; i < dataNum; i++) {
      sprintf(printBuff, "%02x", ud.dataBuff8[i]);
      rxdata += printBuff;
    }
    rxdata.toUpperCase();
    Serial.println(rxdata);

    // restoration of received data
    sec = ud.dataBuff32[1] / 1000;      // transmission time
    Vbatt = ud.dataBuff32F[2];          // battery voltage

    Serial.printf("Slave UID:\t%08x\n", slaveuid);
    Serial.print("Interval:\t"); Serial.print(receivedInterval, 3); Serial.println(" sec");
    Serial.print("Vbatt:\t\t"); Serial.print(Vbatt, 3); Serial.println(" V");

    // RSSI and SNR of the last received packet
    rssi = radio.getRSSI();
    Serial.print("RSSI:\t\t"); Serial.print(rssi); Serial.println(" dBm");
    snr = radio.getSNR();
    Serial.print("SNR:\t\t"); Serial.print(snr); Serial.println(" dB");

    display.clearDisplay();
    display.setCursor(0, 15); display.print("from");
    display.setCursor(64, 15); display.print(slaveuid, HEX);  
    display.setFont(&tahoma5pt7b);
    display.setCursor(0, 31); display.print(rxdata);
    display.setFont(&tahoma7pt7b);
    display.setCursor(10, 47); display.print(rssi); display.print(" dBm");
    display.setCursor(74, 47); display.print(snr); display.print(" dB");
    display.setCursor(10, 63); display.print(receivedTime / 1000); display.print(" sec");
    display.setCursor(74, 63); display.print(Vbatt, 3); display.print(" V");   
    display.display();
    
    // write to SD
    logFile.print(receivedTime); logFile.print(",");
    logFile.print(receivedInterval); logFile.print(",");
    logFile.print(rxdata); logFile.print(",");
    logFile.printf("%08x", slaveuid); logFile.print(",");
    logFile.print(Vbatt, 3); logFile.print(",");
    logFile.print(rssi); logFile.print(",");
    logFile.println(snr);
    logFile.flush();

    delay(1000); // for display
  }
  else {
    // some other error occurred when receiving
    Serial.print("receive failed, code ");
    Serial.println(state);
    errorBlink_1(3);                        // error [[ 3 ]]                  
  }  


  // Master transmits response data to Slave  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Serial.println("Transmitting response packet ... ");

  // send data packet
  ud.dataBuff32[0] = masteruid;         // Master device UID
  ud.dataBuff16[2] = rssi;              // Master RSSI  [dBm]
  ud.dataBuff16[3] = snr;               // Master SN ratio  [dB]
  ud.dataBuff32F[2] = receivedInterval; // packet reserved interval [sec]

  digitalWrite(LED_RED, LOW);
  transmittedFlag = false;
  state = radio.startTransmit(ud.dataBuff8, dataNum);
  digitalWrite(LED_RED, HIGH);

  // print the packet
  txdata = "";
  for(int i = 0; i < dataNum; i++) {
    sprintf(printBuff, "%02x", ud.dataBuff8[i]);
    txdata += printBuff;
  }
  txdata.toUpperCase();
  Serial.println(txdata);

  Serial.printf("Master UID:\t%08x\n", masteruid);
  Serial.print("RSSI\t\t"); Serial.print(rssi); Serial.println(" dBm");
  Serial.print("SNR\t\t"); Serial.print(snr); Serial.println(" dB");

  display.clearDisplay();
  display.setCursor(0, 15); display.print("Master");
  display.setCursor(64, 15); display.print(masteruid, HEX);  
  display.setFont(&tahoma5pt7b);
  display.setCursor(0, 31); display.print(txdata);
  display.setFont(&tahoma7pt7b);
  display.setCursor(10, 47); display.print(rssi); display.print(" dBm");
  display.setCursor(74, 47); display.print(snr); display.print(" dB");
  display.setCursor(10, 63); display.print(receivedInterval, 1); display.print(" sec");
  display.setCursor(74, 63); display.print(Vbatt, 3); display.print(" V");
  display.display();   
   
  // wait for transmittion completion
  while(!transmittedFlag) {
    delay(1);   // wait for SPI ready
  }

  // error status check
  if (state == RADIOLIB_ERR_NONE) {
    // packet was successfully sent
    Serial.println("transmission finished!");
    Serial.println();
  } 
  else {
    // some other error occurred
    Serial.print("failed, code ");
    Serial.println(state);
  }
}

//*********************************************************************************
// callback function when received or transmitted
void setFlag(void) 
{
  uint16_t irqstatus = radio.getIrqStatus();
  if(irqstatus == RADIOLIB_SX126X_IRQ_RX_DONE) {
    receivedFlag = true;
  }
  else if(irqstatus == RADIOLIB_SX126X_IRQ_TX_DONE) {
    transmittedFlag = true;
  }
  else {
    receivedFlag = false;
    transmittedFlag = false;
  }
}

//*********************************************************************************
// Blinks specified number of times on error
void errorBlink(uint8_t err)
{
  while(true) {
    for(int i = 0; i < err; i++) {
      digitalWrite(LED_RED, LOW);
      delay(5);
      digitalWrite(LED_RED, HIGH);
      delay(300);
    }
  delay(500);
  }
}

void errorBlink_1(uint8_t err)
{
  for(int i = 0; i < err; i++) {
    digitalWrite(LED_RED, LOW);
    delay(5);
    digitalWrite(LED_RED, HIGH);
    delay(300);
  }
  delay(500);
}